

(* The output of this is a forest, with one tree per line.  Each tree is like
	the output of learnDT but with slashes (/) instead of newlines separating the
	nodes.  I guess it's not technically a csv, but we'll pretend it is.  :)  *)

Random.self_init ();;

(* type definitions *)
(* leaf: nodenum value *)
(* split: nodenum feature, splitval, left, right *)
type decision_tree = Leaf of int * int | Split of int * int * float * decision_tree * decision_tree;;


(* Give me a filename and I'll give you a list of lists of values (as strings). *)
let read_csv filename =
	let rec input_lines file =
		match try [input_line file] with End_of_file -> [] with
		[] -> []
		| line -> (Str.split (Str.regexp_string ",") (List.nth line 0)) :: input_lines file in
	let file = open_in filename in
	input_lines file;;

(* converts data (a list of lists of strings) to a list of lists of floats *)
let floaterize data =
	let floaterize_elem elem =
		float_of_string elem in
	let floaterize_row row =
		List.map floaterize_elem row in
	List.map floaterize_row data;;
	
let rec extract_feature data feature =
	let func row = List.nth row feature in
	List.map func data;;

(* returns true if all elements of list are quals; false otherwise *)
let all_equal list =
	let first = (List.hd list) in
	let does_not_equal_first item =
		item <> first in
	let unequal_to_first = List.filter does_not_equal_first list in
	if (List.length unequal_to_first) == 0 then true else false;;
	
(* data is a list of lists of values, feature is the feature we want the median of *)
let median data feature =
	let values_sorted = List.sort compare (extract_feature data feature) in
	let index = ((List.length values_sorted)+1)/2 - 1 in
	List.nth values_sorted index;;

(* this isn't the most efficient way to do this, what with iterating through the list twice, but oh well *)
(* also, it is not recursive in the general case; only if it encounters a specific problem *)
let rec split data feature splitval =
	let left_func row = if (List.nth row feature) <= splitval then true else false in
	let right_func row = if (List.nth row feature) > splitval then true else false in
	let (left, right) = (List.filter left_func data, List.filter right_func data) in
	if (List.length right) == 0
		then split data feature (splitval -. 0.000000001)
		else (left, right);;
	

let features_choices data =
	(* aux returns true if there is some variation on a feature; false if not *)
	let aux feature =
		let values = extract_feature data feature in
		if all_equal values then false else true in
	let result = List.filter aux [1;2;3;4;5;6;7;8;9;10;11;12;13] in
	if (List.length result) == 0
		then failwith "There are some datapoints which differ in their score but not any of their attributes.  You should be ashamed!"
		else result;;
	
(* pick a feature from those in which at least some of the data differ *)
(* OMG THIS IS UGLY AND KLUDGY *)
let pick_feature data =
	let choices = features_choices data in
	List.nth choices (Random.int (List.length choices));; 


let rec last_nodenum tree = 
	match tree with
		Leaf(nodenum, _) -> nodenum
	| Split(nodenum, _, _, _, right) -> last_nodenum right;;

let rec learn_tree trainingexamples starting_nodenum=
	if (List.length trainingexamples) == 1
		then Leaf (starting_nodenum, (int_of_float (List.hd (List.hd trainingexamples))))
		else begin
			let feature = pick_feature trainingexamples in
			let splitval = median trainingexamples feature in
			let (leftexamples, rightexamples) = split trainingexamples feature splitval in
			let lefttree = learn_tree leftexamples (starting_nodenum + 1) in
			let last_left_nodenum = last_nodenum lefttree in
			let righttree = learn_tree rightexamples (last_left_nodenum + 1) in
			Split (starting_nodenum, feature, splitval, lefttree, righttree) end;;

let print_tree_helper (num, typenum, feature, splitval, leftnum, rightnum, leafval) =
	print_int num;        print_string ",";
	print_int typenum;    print_string ",";
	print_int feature;    print_string ",";
	print_float splitval; print_string ",";
	print_int leftnum;    print_string ",";
	print_int rightnum;   print_string ",";
	print_int leafval;    print_string "/";;
	
let nodenum_of tree =
		match tree with
			Leaf (nodenum, _) -> nodenum
		| Split (nodenum, _, _, _, _) -> nodenum;;

let rec print_tree tree =
	match tree with
		Leaf (nodenum, value) ->
			print_tree_helper (nodenum, 0, 0, 0.0, 0, 0, value)
	| Split (nodenum, feature, splitval, left, right) -> begin
			print_tree_helper (nodenum, 1, feature, splitval, (nodenum_of left), (nodenum_of right), 0);
			print_tree left;
			print_tree right end;;
			
let rec learn_forest data n =
	match n with
		0 -> []
	| _ -> (learn_tree data 0) :: (learn_forest data (n-1));; 
	
(* forest is a list of trees *)
let rec print_forest forest =
	match forest with
		[] -> ()
	| tree :: rest -> begin
			print_tree tree;
			print_endline "";
			print_forest rest end;;
			


(* Check that there is precisely one argument. *)
let num_args = Array.length Sys.argv in
if num_args != 3 then failwith "Requires two arguments, some training data in CSV format and the number of trees you want in the forest.";;

(* Check that the argument points to a real file. *)
let ensure_exists filename =
	if not (Sys.file_exists filename) then failwith ("The file " ^ filename ^ " does not exist.") in
ensure_exists Sys.argv.(1);;

let data_as_strings = read_csv Sys.argv.(1) in
let data = floaterize data_as_strings in
let forest_size = int_of_string Sys.argv.(2) in
let forest = learn_forest data forest_size in
print_forest forest;;